; PP3X test EPROM v2.1
;
; written by Martin Mansfield 2017-19
;
scratchram	.equ    2000h
ramsize		.equ    0800h
ramtestval	.equ	55h

eeprom		.equ	2800h
eesize		.equ	0800h

pagedram	.equ	4000h
pagesize	.equ	4000h
maxpages	.equ	32

; global variables
globalvars	.equ	scratchram

; offsets for iy access
dsplistoff	.equ	0
keyrowmskoff	.equ	8		; bit mask for row of current key
keyrowidxoff	.equ	9		; row value for current key
keycolidxoff	.equ	10		; column value for current key
keycntoff	.equ	11		; scan count for current key
keyflgoff	.equ	12		; key detected flag for current key
rowmskoff	.equ	13
colidxoff	.equ	14
colmskoff	.equ	15
colmskoff2	.equ	16
scancntoff	.equ	17		; count of scans while no key pressed
scancntoff2	.equ	18
;
uarterroff	.equ	19		; set if uart missing/faulty
nmiflagoff	.equ	20		; set if NMI detected

globalvarsize	.equ	21

; absolute addresses
dsplist		.equ	globalvars+dsplistoff	; 8 bytes storage for 8255 control bytes
nmiflag		.equ	globalvars+nmiflagoff
tmpbuf		.equ	globalvars+globalvarsize

; base address of 6551 uart
uartbase	.equ	3000h
; reg offsets
uartdata	.equ	0
uartsts		.equ	1
uartcmd		.equ	2
uartctl		.equ	3

uarttxbit	.equ	4

; control register value for 19200,8,n,1
ctlval		.equ	1fh
; command register value for no parity, no interrupts
cmdval		.equ	0bh
; command register test values
testwval	.equ	0efh
testrval	.equ	0e0h

; base address of 8255 ports
port1base	.equ	3800h	; IC33 PPI1
port2base	.equ	3810h	; IC32 PPI2
port3base	.equ	3820h	; IC34 PPI3
port4base	.equ	3830h	; IC1  PPI4

; reg offsets
ppiporta		.equ	0
ppiportb		.equ	1
ppiportc		.equ	2
ppicontrol		.equ	3

; control reg values
port1ctlval	.equ	88h	; mode 0 ports a & b output, port c upper input, lower output

port2ctlval	.equ	92h	; mode 0 ports a & b input, port c output
port2beepon	.equ	0fh	; turn on beeper
port2beepoff	.equ	0eh	; turn off beeper
port2waglon	.equ	0dh	; turn on WAGL
port2wagloff	.equ	0ch	; turn off WAGL
port2wagron	.equ	0bh	; turn on WAGR
port2wagroff	.equ	0ah	; turn off WAGR

port3ctlval	.equ	9bh	; mode 0 ports a, b & c input

port4ctlval	.equ	91h	; mode 0 port a input, port b output, port c upper output, lower input
port4dadon	.equ	0fh	; data output on
port4dadoff	.equ	0eh	; data output off
port4cldon	.equ	0dh	; clock output on
port4cldoff	.equ	0ch	; clock output off
port4rsdon	.equ	0bh	; enable power on reset
port4rsdoff	.equ	0ah	; disable power on reset
port4pc4on	.equ	09h	; PC4 high
port4pc4off	.equ	08h	; PC4 low

; port bit values
port1aedg	.equ	01h	; ?
port1acon	.equ	04h	; voltage source control
port1ar28	.equ	10h	; ?
port1ar24	.equ	20h	; ?
port1al24	.equ	40h	; ?
port1al28	.equ	80h	; ?

; display constants
dspdatabits	.equ	8
dspposval	.equ	0a0h
spacechar	.equ	20h
crchar		.equ	0dh
lfchar		.equ	0ah
bytechsize	.equ	2	; characters used to display a byte
addrchsize	.equ	4	; characters used to display an address
mempgpos	.equ	14	; display pos for memory page
memfailpos	.equ	2	; display pos for failure info
selprmtpos	.equ	10	; display pos for select options
optnumpos	.equ	12	; display pos for max option
daclevpos	.equ	10	; display pos for dac output level
dacenpos	.equ	14	; display pos for dac output enable

; keyboard constants
keycnt		.equ	20
scanloopval	.equ	256	; loop value while no key pressed

beepcnt		.equ	200
beeplen		.equ	50	; 1 sec
chirplen	.equ	5	; 0.1 sec

us100dlval	.equ	31	; loop value for 100us delay
sec1dlval	.equ	22472	; loop value for 1 sec delay

; base address of ATL/ATR ?
atlbase		.equ	3840h
atrbase		.equ	3850h

; base addresses of DACs
dac1base	.equ	3900h
dac2base	.equ	3910h
dac3base	.equ	3920h

; base address of EDC/BFE ?
edcbase		.equ	3960h
bfebase		.equ	3970h

.org 0000h
reset
; set up stack pointer for NMI routine
		ld	sp,scratchram+ramsize-1
		jp	start
.org 0008h
		jp	stop
.org 0010h
		jp	stop
.org 0018h
		jp	stop
.org 0020h
		jp	stop
.org 0028h
		jp	stop
.org 0030h
		jp	stop
.org 0038h
		jp	stop

stop		halt

;
; if any of these loops are called then while M1 is active
; the address bits 2 & 3 will indicate the failing PPI IC
; and the chip select for that IC can be probed
;
; A3 A2
; 0  0     PPI 1
; 0  1     PPI 2
; 1  0     PPI 3
; 1  1     PPI 4
;
.org 0040h
ppi1flp		ld	a,(hl)		; loop to indicate fault in PPI 1
		jr	ppi1flp
.org 0044h
ppi2flp		ld	a,(hl)		; loop to indicate fault in PPI 2
		jr	ppi2flp
.org 0048h
ppi3flp		ld	a,(hl)		; loop to indicate fault in PPI 3
		jr	ppi3flp
.org 004ch
ppi4flp		ld	a,(hl)		; loop to indicate fault in PPI 4
		jr	ppi4flp

.org 0066h
;
; NMI indicates problem with programmable voltage sources
;
		push	af
		xor	a
		ld	(dac1base),a	; set voltage sources to 0V
		ld	(dac2base),a
		ld	(dac3base),a
		inc	a
		ld	(nmiflag),a	; set NMI flag
		pop	af
		retn
	
; keycode maps
; hex digits 0 - F
digitmap	.db	15h, 25h, 24h, 29h, 35h, 34h, 39h, 45h, 44h, 49h, 21h, 22h, 31h, 32h, 41h, 42h, 0
; down, up, left , right
cursormap	.db	14h, 19h, 11h, 12h, 0
; load, input, output, verify, hidden1, hidden2, hidden3, empty, edit, list, set, hidden4, csum, hidden5, exit, program
cmdmap		.db	48h, 47h, 46h, 43h, 38h, 37h, 36h, 33h, 28h, 27h, 26h, 23h, 18h, 17h, 16h, 13h, 0

cmdstop		.equ	14

versionstr	.db	"PP3X TEST V2-1  ",0
dispclr		.db	"                ",0
selprmt		.db	"0-X ?",0
mainprmt	.db	"MAIN MENU       ",0
stdprmt		.db	"STD TESTS       ",0
dbgprmt		.db	"DBG TESTS       ",0
memcmdmsg	.db	"TEST MEM PAGE   ",0
sercmdmsg	.db	"TEST SERIAL     ",0
eecmdmsg	.db	"DUMP EEPROM     ",0
lpcmdmsg	.db	"MEM RFSH TIMING ",0
wrlpcmdmsg	.db	"MEM WR/RD TIMING",0
daccmdmsg	.db	"TEST DACS       ",0
dacprbmsg	.db	"PROBE NOW       ",0
dacminmsg	.db	"MIN",0
dacmidmsg	.db	"MID",0
dacmaxmsg	.db	"MAX",0
dacenmsg	.db	"ON",0
dacdismsg	.db	"  ",0
memfailmsg	.db	"F",0
uartfailmsg	.db	"NO UART",0
message		.db	"ABCDEFGHIJKLMNOPQRSTUVWXYZ ",0
hexlookup	.db	"0123456789ABCDEF"

; menu definitions
menu0		.dw	mainprmt
		.db	emenu0-cmds0/2
		.db	0
cmds0
		.dw	stdmenu		; standard tests
		.dw	dbgmenu		; debug tests
emenu0

menu1		.dw	stdprmt
		.db	emenu1-cmds1/2
		.db	0
cmds1
		.dw	memtest		; paged memory test
		.dw	txmessage	; serial port test
		.dw	eedump		; EEPROM contents dump to serial port
emenu1

menu2		.dw	dbgprmt
		.db	emenu2-cmds2/2
		.db	0
cmds2
		.dw	looptest	; mem refresh timing test
		.dw	wrtest		; mem write/read timing test
		.dw	dactest		; dac test
emenu2

start
;
; PPI 2 (IC32) test
;
; configure PPI 2 in mode 0 for buzzer
		ld	hl,port2base+ppicontrol
		ld	(hl),port2ctlval

; check that PPI 2 port A is reading sensible values
		ld	a,(port2base+ppiporta)
		cp	0ffh
		jr	nz,ppi2afault
		ld	(hl),port2waglon	; enable pull-downs
		ld	a,(port2base+ppiporta)
		or	a
		jr	z,ppi2atestend
ppi2afault	ld	hl,port2base+ppiporta	; port to read
		jp	ppi2flp			; indicate PPI 2 fault
ppi2atestend	ld	(hl),port2wagloff	; disable pull-downs

; check that PPI 2 port B is reading sensible values
		ld	a,(port2base+ppiportb)
		cp	0ffh
		jr	nz,ppi2bfault
		ld	(hl),port2wagron	; enable pull-downs
		ld	a,(port2base+ppiportb)
		or	a
		jr	z,ppi2btestend
ppi2bfault	ld	hl,port2base+ppiportb	; port to read
		jp	ppi2flp			; indicate PPI 2 fault
ppi2btestend	ld	(hl),port2wagroff	; disable pull-downs


beepreset	ld	b,beeplen	; 1 sec beep
		ld	ix,beepend	; set return address to following code
;
; sound the buzzer
;
beep		ld	hl,port2base+ppicontrol
bplp
; turn on buzzer
		ld	(hl),port2beepon
		ld	c,beepcnt
bplp2		dec	c
		jr	nz,bplp2

; turn off buzzer
		ld	(hl),port2beepoff
		ld	c,beepcnt
bplp3		dec	c
		jr	nz,bplp3

		dec	b
		jr	nz,bplp

		jp	(ix)		; either continue if called at reset or return to caller
beepend

; simple static RAM test before using stack

scramtest	ld	hl,scratchram
		ld	bc,ramsize
		ld	ix,ramtestend

ramtest
rtlp		ld	d,(hl)		; save original
		ld	a,ramtestval
		ld	(hl),a		; store test value
		cp	(hl)
		jr	nz,rterr	; beep loop if error
		cpl
		ld	(hl),a		; store aah
		cp	(hl)
		jr	nz,rterr	; beep loop if error
		ld	(hl),d		; restore contents
		inc	hl
		dec	bc
		ld	a,b	
		or	c
		jr	nz,rtlp
		jr	rtret		; finished

rterr		scf

rtret		jp	(ix)		; either continue if called at reset or return to caller

ramtestend
		jr	c,beepreset	; jump to beep again if error
;
; PPI 1 (IC33) test
;
; configure PPI 1 in mode 0
		ld	hl,port1base+ppicontrol
		ld	(hl),port1ctlval

; check if CTS input on port C is low
		ld	a,(port1base+ppiportc)
		and	10h		; bit 4 low ?
		jr	z,ppi1testend
ppi1fault	ld	hl,port1base+ppiportc	; port to read
		jp	ppi1flp		; indicate PPI 1 fault
ppi1testend

;
; PPI 3 (IC34) test
;
; configure PPI 3 in mode 0
		ld	hl,port3base+ppicontrol
		ld	(hl),port3ctlval

; check that PPI 3 port A is reading sensible values
		ld	hl,port2base+ppicontrol
		ld	a,(port3base+ppiporta)
		cp	0ffh
		jr	nz,ppi3fault
		ld	(hl),port2waglon	; enable pull-downs
		ld	a,(port3base+ppiporta)
		or	a
		jr	z,ppi3testend
ppi3fault	ld	hl,port3base+ppiporta	; port to read
		jp	ppi3flp			; indicate PPI 3 fault
ppi3testend	ld	(hl),port2wagloff	; disable pull-downs

;
; PPI 4 (IC1) test
;
; configure PPI 4 in mode 0
		ld	hl,port4base+ppicontrol
		ld	(hl),port4ctlval
; enable power on reset for display
		ld	(hl),port4rsdon

; check that PPI 4 port A is reading sensible values
		ld	hl,port2base+ppicontrol
		ld	a,(port4base+ppiporta)
		cp	0ffh
		jr	nz,ppi4fault
		ld	(hl),port2wagron	; enable pull-downs
		ld	a,(port4base+ppiporta)
		or	a
		jr	z,ppi4testend
ppi4fault	ld	hl,port4base+ppiporta	; port to read
		jp	ppi4flp			; indicate PPI 4 fault
ppi4testend	ld	(hl),port2wagroff	; disable pull-downs

; set up stack pointer
		ld	sp,scratchram+ramsize-1
; initialise variables
		call	varinit
; initialise UART
		call	uartinit
		jr	nc,uartdone
		ld	a,(globalvars+uarterroff)
		cpl
		ld	(globalvars+uarterroff),a	; set error flag
uartdone
; initialise display and keypad
		call	dspinit

; 1 sec delay
		call	s1delay
; display version
		ld	hl,versionstr
		call	dspmess

		call	s1delay

; display main menu
		ld	hl,menu0
		call	menudisp

; perform soft reset		
resettest	call	keyscanchk	; wait if key still detected
		jr	nc,resettest	; scan again if returns another keypress
		jp	reset

; standard test menu
stdmenu		ld	hl,menu1
		jp	menudisp

; debug test menu
dbgmenu		ld	hl,menu2
		jp	menudisp

; test on-board memory
memtest		ld	hl,memcmdmsg
		call	dspmess		; display message
		xor	a
		ld	c,a
		ld	b,maxpages
memtestlp	ld	hl,port2base+ppiportc
		ld	(hl),c		; select memory page

		push	bc
		ld	b,c
		ld	a,mempgpos
		call	dsphexbyte	; display memory page
		pop	bc

		call	pgramtest
		jr	nc,memtestnext	; failed ?

memtestfail	push	hl		; save address
		push	de
		ld	de,pagedram
		or	a		; clear carry flag
		sbc	hl,de		; convert address to offset
		pop	de

		push	hl
		call	dspclr
		ld	hl,memfailmsg
		call	dspmess
		pop	hl

		ld	b,c		; save memory page
		ld	c,a		; save value last written
		ld	a,memfailpos
		call	dsphexbyte	; display mem page
		add	a,bytechsize+1
		call	dsphexaddr
		add	a,addrchsize+1
		ld	b,c
		call	dsphexbyte
		add	a,bytechsize+1
		pop	hl		; restore address
		ld	b,(hl)		
		call	dsphexbyte
		call	keyscan		; wait for a keypress
		jr	memtestend

memtestnext	inc	c		; no
		djnz	memtestlp

memtestend	ld	(hl),0		; set to default page
		ret

; output a character string to serial port 256 times
txmessage	ld	hl,sercmdmsg
		call	dspmess		; display message
		ld	a,(globalvars+uarterroff)	; check uart flag
		or	a
		jr	z,txmessok	; uart present ?
		call	dspclr
		ld	hl,uartfailmsg	; no
		call	dspmess
		call	s1delay
		jr	txmessend
txmessok
		ld	iy,message
		ld	b,0
txmesslp	call	txstring
		djnz	txmesslp
txmessend	ret

; dump the EEPROM contents to serial port
eedump		ld	hl,eecmdmsg
		call	dspmess		; display message
		ld	a,(globalvars+uarterroff)	; check uart flag
		or	a
		jr	z,eedumpok	; uart present ?
		call	dspclr
		ld	hl,uartfailmsg	; no
		call	dspmess
		call	s1delay
		jr	eedumpend
eedumpok
		ld	hl,eeprom
		ld	bc,eesize
eedumplp	ld	a,(hl)
		call	txhexbyte
		ld	a,spacechar
		call	txchar
		inc	hl
		dec	bc
		ld	a,0fh		; send cr/lf every 16 bytes
		and	c
		jr	nz,nocrlf
		call	txcrlf
nocrlf		ld	a,b
		or	c
		jr	nz,eedumplp
eedumpend	ret

; infinite loop for probing DRAM refresh etc
looptest	ld	hl,lpcmdmsg
		call	dspmess		; display message
looptestlp	jr	looptestlp

; infinite loop for probing DRAM refresh etc
wrtest		ld	hl,wrlpcmdmsg
		call	dspmess		; display message
; select memory page zero
		xor	a
		ld	(port2base+ppiportc),a	; select memory page

		ld	a,55h
		ld	hl,pagedram
wrtestlp	ld	(hl),a
		cp	(hl)
		jr	wrtestlp

; DAC test
dactest		ld	hl,dacprbmsg
		call	dspmess
		call	s10delay

		ld	hl,daccmdmsg
		call	dspmess
		ld	a,0ffh		; set voltage sources to max
		ld	hl,dacmaxmsg
		call	dacset

		ld	a,07fh		; set voltage sources to mid
		ld	hl,dacmidmsg
		call	dacset

		ld	a,00h		; set voltage sources to min
		ld	hl,dacminmsg
		call	dacset
		ret
;
; set the values for the DACs and disable and enable each for 10 seconds
;
; value in acc, pointer to msg in hl
;
dacset		ld	(dac1base),a	; set voltage sources to value
		ld	(dac2base),a
		ld	(dac3base),a
		ld	a,daclevpos
		call	dspmesspos

		call	keychirp
		call	s10delay

		ld	a,dacenpos	; display enabled
		ld	hl,dacenmsg
		call	dspmesspos

		ld	hl,port1base+ppiporta
		ld	(hl),port1acon	; enable sources

		call	keychirp
		call	s10delay

		ld	(hl),0		; disable sources


		ld	a,dacenpos	; remove enabled
		ld	hl,dacdismsg
		call	dspmesspos
		ret	

; *****************************************************************
;
; paged memory test
;
; on return carry clear if passed, carry set if failed
; failing address in hl, value stored in a
;
pgramtest	push	bc
		push	de
		push	ix
		ld	hl,pagedram
		ld	bc,pagesize
		ld	ix,pgramtestret	; provide return address
		jp	ramtest
pgramtestret	pop	ix
		pop	de
		pop	bc
		ret
		
; *****************************************************************
;
; display the current menu
;
; pointer to menu definition in hl
;
menudisp	push	hl
		ld	e,(hl)
		inc	hl
		ld	d,(hl)
		inc	hl

		ld	b,(hl)		; get number of cmds
		inc	hl
		inc	hl		; points to cmd table

		push	hl
		ld	h,d
		ld	l,e

; display menu title
		call	dspmess
; overlay with prompt
		ld	hl,selprmt
		ld	a,selprmtpos
		call	dspmesspos
		pop	hl
; customise prompt
		dec	b		; convert to zero based max value
		ld	a,optnumpos
		call	dsphexchar	; display max option
		inc	b		; restore

; scan the keyboard
menuscanlp	call	keyscan		; wait for a key press

		call	getcmdkeycode	; check for stop key
		jr	c,menucmdval

		cp	cmdstop		; stop key ?
		jr	nz,menuscanlp	; no
		call	keychirp
		jr	menuret		; yes

menucmdval	call	getdigkeycode	; key code in acc
		jr	c,menuscanlp

		cp	b		; valid command ?
		jr	nc,menuscanlp

		call	keychirp
		call	dspclr		; clear display


; call cmd table entry as subroutine
		ld	d,h
		ld	e,l		; save pointer to cmd table
		ld	l,a		; offset in hl
		xor	a
		ld	h,a
		add	hl,hl		; convert to word offset
		add	hl,de
		ld	e,(hl)		; get jump table contents
		inc	hl
		ld	d,(hl)

		ld	hl,menulp	; push return address on stack
		push	hl

		ld	h,d
		ld	l,e
		jp	(hl)		; jump to it

menulp		pop	hl
		jr	menudisp

menuret		pop	hl
		ret

;
; keyboard subroutines
;

;
; *****************************************************************
;
; check if key is digit, cursor or command
;
; key in acc
;
; on return key id in acc with carry clear if valid
;                           or carry set if not
;
getdigkeycode	push	hl
		ld	hl,digitmap
		jr	findkey

getcurkeycode	push	hl
		ld	hl,cursormap
		jr	findkey

getcmdkeycode	push	hl
		ld	hl,cmdmap

findkey		push	bc
		ld	b,a
		ld	c,0
findkeylp	ld	a,(hl)
		or	a
		jr	nz,findkeycmp	; end of list ?
		ld	a,b		; restore key id
		scf			; not found, carry flag set
		jr	findkeyret

findkeycmp	cp	b
		jr	z,findkeyok	; found it
		inc	hl
		inc	c
		jr	findkeylp
;
findkeyok	ld	a,c		; return offset in list
					; carry flag clear
findkeyret	pop	bc
		pop	hl
		ret
;
; initialise variables
;
varinit:
;
	xor	a
	ld	(globalvars+uarterroff),a

	ret
;
; UART subroutines
;

; uartinit
;
; returns with carry clear if ok, set if error
;
uartinit
		push	bc
		ld	b,a
		push	bc

		ld	a,testwval
		ld	(uartbase+uartcmd),a
		xor	a
		ld	(uartbase+uartsts),a	; reset uart
		ld	a,(uartbase+uartcmd)	; check if reset
		cp	testrval
		jr	z,uartok		; uart ok ?
		scf				; no, faulty
		jr	uartret
uartok
; configure data size etc
		ld	a,ctlval
		ld	(uartbase+uartctl),a
		ld	a,cmdval
		ld	(uartbase+uartcmd),a
uartret
		pop	bc
		ld	a,b
		pop	bc
		ret
		
; acc contains char to transmit
txchar		push	af
; wait for transmitter to empty
txchlp		ld	a,(uartbase+uartsts)
		bit	uarttxbit,a
		jr	z,txchlp
		pop	af
		ld	(uartbase+uartdata),a
		ret

; transmit cr/lf
txcrlf		push	af
		ld	a,crchar
		call	txchar
		ld	a,lfchar
		call	txchar
		pop	af
		ret

; acc contains hex value to transmit
txhexbyte	push	af
		srl	a	
		srl	a	
		srl	a	
		srl	a
		call	txhexchar
		pop	af
		and	0fh
		call	txhexchar
		ret

; acc contains hex nibble to transmit
txhexchar	push	de
		push	hl
		ld	hl,hexlookup
		ld	d,0
		ld	e,a
		add	hl,de
		ld	a,(hl)
		pop	hl
		pop	de
		jr	txchar

; register iy contains pointer to string
txstring	push	af
		push	iy
txstringlp	ld	a,(iy)
		or	a
		jr	z,txstrret
		call	txchar
		inc	iy
		jr	txstringlp
txstrret	pop	iy
		pop	af
		ret

; 8255 subroutines

;
; initialise display hardware
;
dspinit
; disable power on reset
		ld	hl,port4base+ppicontrol
		ld	(hl),port4rsdoff
; 100us delay
		call	us100delay
; clear display
		call	dspclr
;
; set display duty cycle to 16/32
;
		ld	a,0f0h
		call	dspsend
;
; init key scan variables
;
		xor	a
		ld	iy,globalvars	; init for variable access
		ld	(iy+keyflgoff),a
		ret
;
; clear display
;
dspclr		push	hl
		ld	hl,dispclr
		call	dspmess
		pop	hl
		ret
;
; display hexadecimal char
;
; number in b
; zero based start position in a
;
dsphexchar	push	hl
		push	de
		push	af
		ld	de,hexlookup
		ld	a,b
		and	0fh		; extract digit
		ld	l,a
		xor	a
		ld	h,a
		add	hl,de
		ld	a,(hl)		; convert to ASCII
		ld	(tmpbuf),a	; store in buffer
		xor	a
		ld	(tmpbuf+1),a	; store null
		pop	af
		pop	de
		ld	hl,tmpbuf
		call	dspmesspos
		pop	hl
		ret
;
; display hexadecimal address
;
; address in hl
; zero based start position in a

dsphexaddr	push	af
		push	bc
		ld	b,h
		call	dsphexbyte
		inc	a
		inc	a
		ld	b,l
		call	dsphexbyte
		pop	bc
		pop	af
		ret
;
; display hexadecimal byte
;
; number in b
; zero based start position in a
;
dsphexbyte	push	hl
		push	de
		push	af
		ld	de,hexlookup
		ld	a,b		; save copy
		srl	a
		srl	a
		srl	a
		srl	a		; extract first digit
		ld	l,a
		xor	a
		ld	h,a
		add	hl,de
		ld	a,(hl)		; convert to ASCII
		ld	(tmpbuf),a	; store in buffer	
		ld	a,b
		and	0fh		; extract second digit
		ld	l,a
		xor	a
		ld	h,a
		add	hl,de
		ld	a,(hl)		; convert to ASCII
		ld	(tmpbuf+1),a	; store in buffer
		xor	a
		ld	(tmpbuf+2),a	; store null
		pop	af
		pop	de
		ld	hl,tmpbuf
		call	dspmesspos
		pop	hl
		ret
;
; display a message on VFD at pos 0
;
; pointer to null terminated message in hl
;
dspmess		push	af
		xor	a
		call	dspmesspos
		pop	af
		ret
;
; display a message on VFD
;
; pointer to null terminated message in hl
; zero based start position in a
;
; PP3X has the display IC chars wired incorrectly from right to left
; so we must size message, adjust start pos and send in reverse order
;
dspmesspos	push	af
		push	bc
		ld	b,0
		ld	c,a		; save pos
dspmslp1	ld	a,(hl)		; size the message
		or	a
		jr	z,dspmspos	; end of message ?
		inc	hl
		inc	b
		jr	dspmslp1
;
dspmspos	ld	a,c		; restore pos
; since the char pos is 15 on left and 0 on right
; the pos must be converted by adding length of message -1
; and using the one's complement of the lower 4 bits
		add	a,b
		dec	a
		cpl
		dec	a		; value is 1 less then zero based offset
		and	0fh

; set character position
		or	dspposval	; build command value
		call	dspsend

dspmssend	dec	hl
		ld	a,(hl)
		and	3fh		; clear top 2 bits
		call	dspsend
		dec	b
		jr	nz,dspmssend	; sent all data ?
		pop	bc
		pop	af
		ret
;
; send byte of data in acc to 10937 display IC
;
dspsend		push	bc
		push	de
		push	hl
		ld	b,dspdatabits
;
; build values to write to port
;
		ld	h,a
		ld	de,dsplist
dspbldlp	ld	a,port4dadoff
		sla	h
		jr	nc,dspsav
		ld	a,port4dadon
dspsav		ld	(de),a		; save value
		inc	de
		djnz	dspbldlp
;
; write values while driving clock output
;
		ld	b,dspdatabits
		ld	hl,port4base+ppicontrol
		ld	de,dsplist
		ld	a,(de)		; moved out of loop when clock high
clock		ld	(hl),port4cldon	; drive clock high
		ld	(hl),a
		nop			; ensure data setup time
		ld	(hl),port4cldoff	; drive clock low
		inc	de
		ld	a,(de)		; moved out of loop when clock high
		ex	(sp),hl		; ensure clock rate slow enough
		ex	(sp),hl
		ex	(sp),hl
		ex	(sp),hl
		djnz	clock
; 100us delay
		call	us100delay

		pop	hl
		pop	de
		pop	bc
		ret
;
; scan the keyboard
;
; keyscan - wait until key pressed
;           keycode returned in reg A, top nibble is row, bottom nibble is column
;
; keyscanchk - wait a fixed time for key state change
;            returns with carry set if no state change
;
keyscanchk	push	de
		ld	de,scanloopval
		jr	keyscancnt

keyscan		push	de
		ld	de,0

keyscancnt	push	bc
		push	iy
		ld	iy,globalvars	; init for variable access
		ld	(iy+scancntoff),e
		ld	(iy+scancntoff+1),d	; store scanloop value

keyscanlp	xor	a			; init scan values
		ld	(iy+colidxoff),a	; set column var to 0
		ld	(iy+colmskoff2),a	; set to 0
		inc	a
		ld	(iy+colmskoff),a	; set to 1

keyscanlp1	ld	a,(iy+colmskoff2)	; update bit 8
		or	a
		jr	nz,pc4on
		ld	a,port4pc4off
		jr	keysetbit
pc4on		ld	a,port4pc4on
keysetbit	ld	(port4base+ppicontrol),a

keyscanlp2	ld	a,(iy+colmskoff)	; update bits 0-7
		ld	(port4base+ppiportb),a
;
; read lower 4 bits of port c
;
		ld	a,(port4base+ppiportc)
		and	0fh		; mask out upper bits
		ld	(iy+rowmskoff),a	; save row value

		ld	a,(iy+keyflgoff)
		or	a		; is there a current key ?
		jr	z,keyfirst	; no
		ld	a,(iy+keycolidxoff)	; yes
		cp	(iy+colidxoff)		; in this column ?
		jr	nz,keycont	; no, ignore it

; key press already detected in this col
keynext		ld	a,(iy+rowmskoff)	; restore row value
		or	a		; key pressed in current col ?
		jr	nz,keydn2	; yes
		jr	keyup1

; no key press already detected
keyfirst	ld	a,(iy+rowmskoff)	; restore row value
		or	a		; key pressed in current col ?
		jr	nz,keydn1	; yes
		jr	keycont

; check for multiple keys pressed
keydn2		ld	a,(iy+keyrowmskoff)
		cp	(iy+rowmskoff)	; the only key pressed ?
		jr	z,keydn3	; yes
		and	(iy+rowmskoff)	; still pressed ?
		jr	nz,keycont	; yes, wait until only key pressed

; detected key up
keyup1		ld	a,(iy+keycntoff)
		cp	keycnt << 1		; has count reached high limit ?
		jr	z,keyup2		; yes
		inc	(iy+keycntoff)		; no, increment count
		jr	keycont

keyup2		xor	a
		ld	(iy+keyflgoff),a	; clear flag
		jr	keycont

;
; detected first key down
;
; convert row mask to row index
keydn1		ld	b,1
		ld	a,(iy+rowmskoff)
keydn1lp	srl	a
		jr	c,keydownchk
		inc	b
		jr	keydn1lp
;
; check for multiple keys
;
keydownchk	or	a		; only one key pressed in this col ?
		jr	nz,keycont	; no, so ignore it
; set variables
		ld	(iy+keycntoff),keycnt	; save count for key
		ld	a,(iy+colidxoff)
		ld	(iy+keycolidxoff),a	; save column for key
		ld	a,(iy+rowmskoff)
		ld	(iy+keyrowmskoff),a	; save row mask for key
		ld	(iy+keyrowidxoff),b	; save row index for key
		ld	(iy+keyflgoff),1	; set flag
		jr	keycont

;
; detected following key down
;
keydn3		ld	a,(iy+keycntoff)
		or	a			; already detected ?
		jr	z,keycont		; yes, wait for it to be released
		dec	(iy+keycntoff)		; decrement count for key
		jr	nz,keycont
;
; build key code from row and column
;	
keycode		ld	a,(iy+keyrowidxoff)
		sla	a
		sla	a
		sla	a
		sla	a		; upper 4 bits are row
		or	(iy+keycolidxoff)	; lower 4 bits are column
		inc	a		; column index based from 1
		jr	keyscanret
;
; continue scan
;
keycont		inc	(iy+colidxoff)	; inc column var
		ld	a,(iy+colmskoff)	; get column mask
		or	a
		jr	z,keychk8	; shift bit 8
		sla	(iy+colmskoff)	; scan next column
		jr	nc,keyscanloop	; bit 8 set ?
		ld	a,80h
		ld	(iy+colmskoff2),a	; init mask
		jr	keyscanloop

;
; deal with bit 8
;		
keychk8		sla	(iy+colmskoff2)	; scan next column
		jr	nc,keyscanloop	; last column ?
		ld	a,(iy+keyflgoff)	; yes, current key ?
		or	a
		jr	nz,keyendlp	; yes, continue scan

keycntchk
		ld	d,(iy+scancntoff+1)	; check if scan count in use
		ld	e,(iy+scancntoff)
		ld	a,d		; is count zero ?
		or	e
		jr	z,keyendlp	; yes, so don't decrement it

; decrement scan count and check for zero
keydeccnt	dec	de
		ld	(iy+scancntoff+1),d
		ld	(iy+scancntoff),e
		ld	a,d
		or	e
		jr	nz,keyendlp	; zero ?
keyout		scf			; yes, set carry and return

keyscanret	pop	iy
		pop	bc
		pop	de
		ret

keyendlp	jp	keyscanlp

keyscanloop	jp	keyscanlp1

;
; chirp when key pressed
;
keychirp	push	af
		push	bc
		push	hl
		push	ix

		ld	b,chirplen
		ld	ix,keychirpret	; provide return address
		jp	beep		; call beep routine as subroutine

keychirpret	pop	ix
		pop	hl
		pop	bc
		pop	af
		ret

; 100us delay
us100delay	push	bc
		ld	b,us100dlval

dllp1		djnz	dllp1
		pop	bc
		ret

; 1s delay
s1delay		push	af
		push	bc
		ld	bc,sec1dlval
s1dllp1		dec	bc
		ex	(sp),hl		; just to waste some time !
		ex	(sp),hl
		ld	a,b
		or	c
		jr	nz,s1dllp1
		pop	bc
		pop	af
		ret

; 10s delay
s10delay	push	bc
		ld	b,10

s10dllp1	call	s1delay
		djnz	s10dllp1
		pop	bc
		ret
.end
